;;; github.el --- Fake package for testing dh-make-elpa  -*- lexical-binding: t; -*-

;; NOTICE: The copyright information below is for testing purpose and does not
;; take any effect.

;; Copyright (C) 2023 by Richard Lewis <richard.lewis.debian@googlemail.com>
;; Copyright (C) 2025 by DhMake Elpa <dh-make-elpa@test.debian>

;; Author: DhMake Elpa <dh-make-elpa@test.debian>
;; Version: 0.10.21
;; URL: https://example.com/home/page

;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 2 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

;;; Commentary:

;; A fake package for testing dh-make-elpa based on an earlier example by
;; Richard Lewis.

;;; Code:
(provide 'github)
